@@ -1,2 +1,2 @@
-from .console import console
+from .console import app
 from .__version__ import __version__
