@@ -38,16 +38,19 @@ from click import pass_context
 from click import secho
 from click import style
 from click.shell_completion import CompletionItem
+from falocalrepo_database import Column
 from falocalrepo_database import Cursor
 from falocalrepo_database import Database
 from falocalrepo_database import Table
 from falocalrepo_database import __version__ as __database_version__
 from falocalrepo_database.database import query_to_sql
 from falocalrepo_database.selector import SelectorBuilder as Sb
+from falocalrepo_database.tables import CommentsColumns
 from falocalrepo_database.tables import HistoryColumns
 from falocalrepo_database.tables import JournalsColumns
 from falocalrepo_database.tables import SubmissionsColumns
 from falocalrepo_database.tables import UsersColumns
+from falocalrepo_database.tables import comments_table
 from falocalrepo_database.tables import journals_table
 from falocalrepo_database.tables import submissions_table
 from falocalrepo_database.tables import users_table
@@ -89,6 +92,7 @@ class TableChoice(CompleteChoice):
         CompletionItem(submissions_table),
         CompletionItem(journals_table),
         CompletionItem(users_table),
+        CompletionItem(comments_table),
     ]
 
 
@@ -129,6 +133,8 @@ class SortColumnsChoice(ColumnsChoice):
           for c in JournalsColumns.as_list()],
         *[(users_table, CompletionItem(c.name, help=f"{users_table}:{c.name}"))
           for c in UsersColumns.as_list()],
+        *[(comments_table, CompletionItem(c.name, help=f"{comments_table}:{c.name}"))
+          for c in CommentsColumns.as_list()],
     ]
 
 
@@ -138,6 +144,7 @@ class SearchColumnsChoice(ColumnsChoice):
         (submissions_table, CompletionItem("@", help=f"{submissions_table}:ALL")),
         (journals_table, CompletionItem("@", help=f"{journals_table}:ALL")),
         (users_table, CompletionItem("@", help=f"{users_table}:ALL")),
+        (comments_table, CompletionItem("@", help=f"{comments_table}:ALL")),
     ]
 
 
@@ -161,6 +168,8 @@ def column_callback(ctx: Context, param: Option, value: tuple[str]) -> tuple[str
         table_columns = [c.name for c in JournalsColumns.as_list()]
     elif table == users_table:
         table_columns = [c.name for c in UsersColumns.as_list()]
+    elif table == comments_table:
+        table_columns = [c.name for c in CommentsColumns.as_list()]
     if (col := next((c for c in value if c not in table_columns and c != "@"), None)) is not None:
         raise BadParameter(f"{col!r} is not one of {', '.join(map(repr, table_columns))}.", ctx, param)
     return value
@@ -203,6 +212,8 @@ def get_table(db: Database, table: str) -> Table:
         return db.submissions
     elif table == journals_table.lower():
         return db.journals
+    elif table == comments_table.lower():
+        return db.comments
     else:
         return db[table]
 
@@ -269,8 +280,8 @@ def print_json(results: Cursor, file: TextIO) -> int:
     return results_total
 
 
-def search(table: Table, headers: list[str], query: str, sort: tuple[tuple[str, str]], limit: int | None,
-           offset: int | None, sql: bool) -> tuple[Cursor, tuple[str, list[str]]]:
+def search(table: Table, headers: list[str | Column], query: str, sort: tuple[tuple[str, str]], limit: int = 0,
+           offset: int = 0, sql: bool = False) -> tuple[Cursor, tuple[str, list[str]]]:
     cols_table: list[str] = [c.name for c in table.columns]
     query_elems, values = ([query], []) if sql or not query.strip() else query_to_sql(
         query, "any", [*map(str.lower, {*cols_table, "any"} - {"ID", "AUTHOR", "USERNAME"})],
@@ -290,6 +301,68 @@ def html_to_ansi(html: str, *, root: bool = False) -> str:
         br.replaceWith("\n")
     for hr in html_parsed.select("hr"):
         hr.replaceWith(f"\n{'-' * (width or 40):^{(width or 1) - 1}}\n")
+    for tag in html_parsed.select("i.smilie"):
+        tag_class: Iterable[str] = tag.get("class", [])
+        if "embarrassed" in tag_class:
+            tag.replaceWith("😃")
+        elif "tongue" in tag_class:
+            tag.replaceWith("😛")
+        elif "cool" in tag_class:
+            tag.replaceWith("😎")
+        elif "wink" in tag_class:
+            tag.replaceWith("😉")
+        elif "oooh" in tag_class:
+            tag.replaceWith("😮")
+        elif "smile" in tag_class:
+            tag.replaceWith("🙂")
+        elif "evil" in tag_class:
+            tag.replaceWith("😈")
+        elif "huh" in tag_class:
+            tag.replaceWith("😵‍💫")
+        elif "whatever" in tag_class:
+            tag.replaceWith(":3")
+        elif "angel" in tag_class:
+            tag.replaceWith("😇")
+        elif "badhairday" in tag_class:
+            tag.replaceWith(":badhair:")
+        elif "lmao" in tag_class:
+            tag.replaceWith("😂")
+        elif "cd" in tag_class:
+            tag.replaceWith("💿")
+        elif "crying" in tag_class:
+            tag.replaceWith("😭")
+        elif "dunno" in tag_class:
+            tag.replaceWith("🤨")
+        elif "embarrassed" in tag_class:
+            tag.replaceWith("😳")
+        elif "note" in tag_class:
+            tag.replaceWith("🎁")
+        elif "coffee" in tag_class:
+            tag.replaceWith("🍺")
+        elif "love" in tag_class:
+            tag.replaceWith("❤️")
+        elif "nerd" in tag_class:
+            tag.replaceWith("🤓")
+        elif "note" in tag_class:
+            tag.replaceWith("🎵")
+        elif "derp" in tag_class:
+            tag.replaceWith("🥴")
+        elif "sarcastic" in tag_class:
+            tag.replaceWith("🤨")
+        elif "serious" in tag_class:
+            tag.replaceWith("😐")
+        elif "sad" in tag_class:
+            tag.replaceWith("☹️")
+        elif "sleepy" in tag_class:
+            tag.replaceWith("🥱")
+        elif "teeth" in tag_class:
+            tag.replaceWith("😡")
+        elif "veryhappy" in tag_class:
+            tag.replaceWith("😃")
+        elif "yelling" in tag_class:
+            tag.replaceWith("🤬")
+        elif "zipped" in tag_class:
+            tag.replaceWith("🤐")
     for [tag_name, tag_style] in (("i", italic), ("strong", bold), ("u", underline), ("s", strikethrough)):
         for tag in html_parsed.select(tag_name):
             for child in tag.select("*"):
@@ -298,7 +371,7 @@ def html_to_ansi(html: str, *, root: bool = False) -> str:
     for a in html_parsed.select("a"):
         for child in a.select("*"):
             child.replaceWith(html_to_ansi(str(child)))
-        a.replaceWith(f"[{a.text.strip()}]({a['href']})")
+        a.replaceWith(f"[{text}]({a['href']})" if (text := a.text.strip()) != a["href"] else a["href"])
     for span in html_parsed.select("span.bbcode[style*=color]"):
         for child in span.select("*"):
             child.replaceWith(html_to_ansi(str(child)))
@@ -330,10 +403,11 @@ def html_to_ansi(html: str, *, root: bool = False) -> str:
     return html_parsed.text
 
 
-def view_entry(entry: dict[str, Any], html_fields: list[str], *, raw_html: bool = False) -> str:
+def view_entry(entry: dict[str, Any], html_fields: list[str], exclude_fields: list[str], *,
+               raw_html: bool = False) -> str:
     outputs: list[str] = []
     padding: int = max(map(len, entry.keys()))
-    for field, value in ((k, v) for k, v in entry.items() if k not in html_fields):
+    for field, value in ((k, v) for k, v in entry.items() if k not in html_fields and k not in exclude_fields):
         output = f"{blue}{field:<{padding}}{reset}: " + \
                  (("\n" + (" " * (padding + 2))).join(value) if isinstance(value, (list, set)) else str(value).strip())
         outputs.append(output)
@@ -343,6 +417,16 @@ def view_entry(entry: dict[str, Any], html_fields: list[str], *, raw_html: bool
     return "\n".join(outputs)
 
 
+def view_comments(comments: list[dict], *, raw_html: bool = False) -> str:
+    outputs: list[str] = []
+    for comment in comments:
+        del comment[CommentsColumns.PARENT_TABLE.name]
+        del comment[CommentsColumns.PARENT_ID.name]
+        outputs.extend([view_entry(comment, [CommentsColumns.TEXT.name], ["REPLIES"], raw_html=raw_html),
+                        view_comments(comment.get("REPLIES", []))])
+    return "\n\n".join(filter(bool, outputs))
+
+
 @group("database", cls=CustomHelpColorsGroup, short_help="Operate on the database.", no_args_is_help=True)
 @color_option
 @help_option
@@ -350,6 +434,7 @@ def view_entry(entry: dict[str, Any], html_fields: list[str], *, raw_html: bool
     users_columns="\n    ".join(f" * {c.name:<8} {c.type.__name__}" for c in UsersColumns.as_list()),
     submissions_columns="\n    ".join(f" * {c.name:<11} {c.type.__name__}" for c in SubmissionsColumns.as_list()),
     journals_columns="\n    ".join(f" * {c.name:<10} {c.type.__name__}" for c in JournalsColumns.as_list()),
+    comments_columns="\n    ".join(f" * {c.name:<12} {c.type.__name__}" for c in CommentsColumns.as_list()),
     prog_name=__prog_name__, version=__version__)
 def database_app():
     """
@@ -369,6 +454,10 @@ def database_app():
     \b
     {cyan}Journals{reset}
     {journals_columns}
+
+    \b
+    {cyan}Comments{reset}
+    {comments_columns}
     """
     pass
 
@@ -406,6 +495,8 @@ def database_info(ctx: Context, database: Callable[..., Database]):
     echo(f"{yellow}{len(db.submissions)}{reset}", color=ctx.color)
     echo(f"{blue}Journals{reset}   : ", nl=False, color=ctx.color)
     echo(f"{yellow}{len(db.journals)}{reset}", color=ctx.color)
+    echo(f"{blue}Comments{reset}   : ", nl=False, color=ctx.color)
+    echo(f"{yellow}{len(db.comments)}{reset}", color=ctx.color)
     echo(f"{blue}History{reset}    : ", nl=False, color=ctx.color)
     echo(f"{yellow}{len(db.history)}{reset}", color=ctx.color)
 
@@ -545,7 +636,7 @@ def database_search(ctx: Context, database: Callable[..., Database], table: str,
 
     db: Database = database()
     db_table: Table = get_table(db, table)
-    headers: list[tuple[str, int]] = [*zip(map(str.upper, column), (*table_widths, *([0] * len(column))))]
+    headers: list[tuple[str | Column, int]] = [*zip(map(str.upper, column), (*table_widths, *([0] * len(column))))]
 
     if table in (submissions_table, journals_table):
         headers = headers or [*zip(cols := [SubmissionsColumns.ID.name, SubmissionsColumns.AUTHOR.name,
@@ -557,16 +648,23 @@ def database_search(ctx: Context, database: Callable[..., Database], table: str,
             *zip(cols := [UsersColumns.USERNAME.name, UsersColumns.ACTIVE.name, UsersColumns.FOLDERS.name],
                  ((*table_widths, *([0] * len(cols))) if table_widths else (40, 6, 0)))]
         sort = sort or ((UsersColumns.USERNAME.name, "ASC"),)
+    elif table == comments_table:
+        headers = headers or [*zip(cols := [CommentsColumns.ID.name, CommentsColumns.PARENT_TABLE.name,
+                                            CommentsColumns.PARENT_ID.name, CommentsColumns.AUTHOR.name,
+                                            CommentsColumns.DATE.value],
+                                   ((*table_widths, *([0] * len(cols))) if table_widths else (10, 11, 10, 16, 0)))]
+        sort = sort or ((CommentsColumns.ID.name, "desc"),)
 
     if any(h == "@" for h, _ in headers):
-        headers = list(zip([c.name for c in db_table.columns], (*table_widths, *([0] * len(db_table.columns)))))
+        headers = list(zip([c for c in db_table.columns], (*table_widths, *([0] * len(db_table.columns)))))
 
     headers[-1] = (headers[-1][0], 0)
     results, [query, values] = search(db_table, [h for h, _ in headers], query, sort, limit, offset, sql)
     results_total: int = 0
 
     if output == Output.table:
-        results_total = print_table(ctx, results, headers, ignore_width)
+        results_total = print_table(ctx, results, [(h.name if isinstance(h, Column) else h, w) for h, w in headers],
+                                    ignore_width)
     elif output == Output.csv:
         results_total = print_csv(results, stdout, ",")
     elif output == Output.tsv:
@@ -587,12 +685,14 @@ def database_search(ctx: Context, database: Callable[..., Database], table: str,
 @argument("table", nargs=1, required=True, is_eager=True, type=TableChoice())
 @argument("id_", metavar="ID", nargs=1, required=True, type=str, callback=id_callback)
 @option("--raw-content", is_flag=True, default=False, help="Do not format HTMl fields.")
+@option("--view-comments", "view_comments_", is_flag=True, default=False,
+        help="Show comments for submissions and journals.")
 @database_exists_option
 @color_option
 @help_option
 @pass_context
 def database_view(ctx: Context, database: Callable[..., Database], table: str, id_: tuple[str | int, ...],
-                  raw_content: bool):
+                  raw_content: bool, view_comments_: bool):
     """
     View a single entry in the terminal. Submission descriptions, journal contents, and user profile pages are rendered
     and formatted.
@@ -609,13 +709,21 @@ def database_view(ctx: Context, database: Callable[..., Database], table: str, i
     if not (entry := get_table(db, table)[id_[0]]):
         secho(f"Entry {id_[0]!r} could not be found in {table.lower()}", fg="red", color=ctx.color)
     elif table == submissions_table:
-        echo(view_entry(entry, [SubmissionsColumns.DESCRIPTION.name], raw_html=raw_content), color=ctx.color)
+        echo(view_entry(entry, [SubmissionsColumns.DESCRIPTION.name], [], raw_html=raw_content), color=ctx.color)
+        if view_comments_:
+            echo(f"\n\n{blue}COMMENTS{reset}:")
+            echo(view_comments(db.comments.get_comments_tree(submissions_table, id_[0]), raw_html=raw_content))
     elif table == journals_table:
-        echo(view_entry(entry, [JournalsColumns.CONTENT.name], raw_html=raw_content), color=ctx.color)
+        echo(view_entry(entry, [JournalsColumns.CONTENT.name], [], raw_html=raw_content), color=ctx.color)
+        if view_comments_:
+            echo(f"\n\n{blue}COMMENTS{reset}:")
+            echo(view_comments(db.comments.get_comments_tree(journals_table, id_[0]), raw_html=raw_content))
     elif table == users_table:
-        echo(view_entry(entry, [UsersColumns.USERPAGE.name], raw_html=raw_content), color=ctx.color)
+        echo(view_entry(entry, [UsersColumns.USERPAGE.name], [], raw_html=raw_content), color=ctx.color)
+    elif table == comments_table:
+        echo(view_entry(entry, [CommentsColumns.TEXT.name], [], raw_html=raw_content), color=ctx.color)
     else:
-        echo(view_entry(entry, [], raw_html=raw_content), color=ctx.color)
+        echo(view_entry(entry, [], [], raw_html=raw_content), color=ctx.color)
 
 
 @database_app.command("remove", no_args_is_help=True, short_help="Remove entries.")
@@ -829,8 +937,9 @@ def database_copy(ctx: Context, database: Callable[..., Database], database_dest
     cursors: list[Cursor]
     if query:
         cursors = [
-            search(tb := get_table(db, t), [c.name for c in tb.columns], q, ((tb.key.name, ""),), None, None, False)[0]
-            for t, q in query]
+            search(tb := get_table(db2, t), tb.columns, q if q.strip() != "%" else "", ((tb.key.name, ""),))[0]
+            for t, q in query
+        ]
     else:
         cursors = [get_table(db, t.value).select() for t in TableChoice.completion_items]
 
@@ -873,8 +982,9 @@ def database_merge(ctx: Context, database: Callable[..., Database], database_ori
     cursors: list[Cursor]
     if query:
         cursors = [
-            search(tb := get_table(db2, t), [c.name for c in tb.columns], q, ((tb.key.name, ""),), None, None, False)[0]
-            for t, q in query]
+            search(tb := get_table(db2, t), tb.columns, q if q.strip() != "%" else "", ((tb.key.name, ""),))[0]
+            for t, q in query
+        ]
     else:
         cursors = [get_table(db2, t.value).select() for t in TableChoice.completion_items]
 
