@@ -65,6 +65,9 @@ output_option = option("--simple-output", is_flag=True, default=False, help="Sim
 dry_run_option = option("--dry-run", is_flag=True, default=False, help="Fetch entries without modifying database.")
 verbose_report_option = option("--verbose-report", is_flag=True, default=False, help="Output full report with IDs.")
 report_file_option = option("--report-file", default=None, type=File("w"), help="Write download report to a file.")
+retry_option = option("--retry", metavar="INTEGER", default=1, type=IntRange(1, 5), show_default=True,
+                      help="Retry downloads.")
+comments_option = option("--save-comments", is_flag=True, default=False, help="Save entries' comments.")
 
 
 def users_callback(ctx: Context, param: Option, value: tuple[str]) -> tuple[str]:
@@ -115,11 +118,15 @@ def download_login(ctx: Context, database: Callable[..., Database]):
         ctx.exit(1)
 
 
+# noinspection DuplicatedCode
 @download_app.command("users", short_help="Download users.", no_args_is_help=True)
 @option("--user", "-u", "users", metavar="USER", required=True, multiple=True, type=str, callback=users_callback,
         help="Username.")
 @option("--folder", "-f", "folders", metavar="FOLDER", required=True, multiple=True, type=DownloadFolderChoice(),
         callback=lambda _c, _p, v: sort_set(v), help="Folder to download.")
+@retry_option
+@comments_option
+@option("--replace", is_flag=True, default=False, show_default=True, help="Replace entries already in database.")
 @dry_run_option
 @verbose_report_option
 @report_file_option
@@ -131,19 +138,29 @@ def download_login(ctx: Context, database: Callable[..., Database]):
                             [Folder.watchlist_by.value + f":{yellow}FOLDER{reset}"] +
                             [Folder.watchlist_to.value + f":{yellow}FOLDER{reset}"]))
 def download_users(ctx: Context, database: Callable[..., Database], users: tuple[str], folders: tuple[str],
-                   dry_run: bool, verbose_report: bool, report_file: TextIO | None):
+                   retry: int | None, save_comments: bool, replace: bool, dry_run: bool, verbose_report: bool,
+                   report_file: TextIO | None):
     """
     Download specific user folders, where {yellow}FOLDER{reset} is one of {0}. Multiple {yellow}--user{reset} and
     {yellow}--folder{reset} arguments can be passed. {yellow}USER{reset} can be set to {cyan}@me{reset} to fetch own
     username. {cyan}watchlist-by:{yellow}FOLDER{reset} and {cyan}watchlist-to:{yellow}FOLDER{reset} arguments add the
     specified {yellow}FOLDER{reset}(s) to the new user entries.
 
+    The {yellow}--retry{reset} option enables downloads retries for submission files and thumbnails up to 5 retries.
+
+    The {yellow}--save-comments{reset} option allows saving comments for downloaded submissions and journals. Comments
+    are otherwise ignored.
+
+    If the {yellow}--replace{reset} option is used, existing entries in the database will be updated (favorites are
+    maintained).
+
     The optional {yellow}--dry-run{reset} option disables downloading and saving and simply lists fetched entries.
     Users are not added/deactivated.
     """
     db: Database = database()
     api: FAAPI = open_api(db, ctx)
-    downloader: Downloader = Downloader(db, api, color=ctx.color, dry_run=dry_run)
+    downloader: Downloader = Downloader(db, api, color=ctx.color, comments=save_comments, retry=retry or 0,
+                                        replace=replace, dry_run=dry_run)
     if not dry_run:
         add_history(db, ctx, users=users, folders=folders)
     watchlist_by: list[str] = [f.split(":")[1] for f in folders if f.startswith(Folder.watchlist_by.value)]
@@ -170,6 +187,7 @@ def download_users(ctx: Context, database: Callable[..., Database], users: tuple
             downloader.verbose_report(report_file)
 
 
+# noinspection DuplicatedCode
 @download_app.command("update", short_help="Download new entries for users in database.")
 @option("--user", "-u", "users", metavar="USER", multiple=True, type=str, callback=users_callback,
         help="User to update.")
@@ -179,6 +197,8 @@ def download_users(ctx: Context, database: Callable[..., Database], users: tuple
         help="Number of submissions to find in the database before stopping.")
 @option("--deactivated", is_flag=True, default=False, help="Check deactivated users.")
 @option("--like", is_flag=True, is_eager=True, default=False, help=f"Consider {yellow}USER{reset} to be LIKE queries.")
+@retry_option
+@comments_option
 @dry_run_option
 @verbose_report_option
 @report_file_option
@@ -188,7 +208,9 @@ def download_users(ctx: Context, database: Callable[..., Database], users: tuple
 @pass_context
 @docstring_format(', '.join(c.value for c in UpdateFolderChoice.completion_items))
 def download_update(ctx: Context, database: Callable[..., Database], users: tuple[str], folders: tuple[str], stop: int,
-                    deactivated: bool, like: bool, dry_run: bool, verbose_report: bool, report_file: TextIO | None):
+                    deactivated: bool, like: bool, retry: int | None, save_comments: bool, dry_run: bool,
+                    verbose_report: bool,
+                    report_file: TextIO | None):
     """
     Download new entries using the users and folders already in the database. {yellow}--user{reset} and
     {yellow}--folder{reset} options can be used to restrict the update to specific users and or folders, where
@@ -204,12 +226,18 @@ def download_update(ctx: Context, database: Callable[..., Database], users: tupl
     The {yellow}--like{reset} option enables using SQLite LIKE statements for {yellow}USER{reset} values, allowing to
     select multiple users at once.
 
+    The {yellow}--retry{reset} option enables downloads retries for submission files and thumbnails up to 5 retries.
+
+    The {yellow}--save-comments{reset} option allows saving comments for downloaded submissions and journals. Comments
+    are otherwise ignored.
+
     The optional {yellow}--dry-run{reset} option disables downloading and saving and simply lists fetched entries.
     Users are not added/deactivated.
     """
     db: Database = database()
     api: FAAPI = open_api(db, ctx)
-    downloader: Downloader = Downloader(db, api, color=ctx.color, dry_run=dry_run)
+    downloader: Downloader = Downloader(db, api, color=ctx.color, comments=save_comments, retry=retry or 0,
+                                        dry_run=dry_run)
     if not dry_run:
         add_history(db, ctx, users=users, folders=folders, stop=stop)
     try:
@@ -225,10 +253,13 @@ def download_update(ctx: Context, database: Callable[..., Database], users: tupl
             downloader.verbose_report(report_file)
 
 
+# noinspection DuplicatedCode
 @download_app.command("submissions", short_help="Download single submissions.", no_args_is_help=True)
 @argument("submission_id", nargs=-1, required=True, type=IntRange(1),
           callback=lambda _c, _p, v: sorted(set(v), key=v.index))
 @option("--replace", is_flag=True, default=False, show_default=True, help="Replace submissions already in database.")
+@retry_option
+@comments_option
 @dry_run_option
 @verbose_report_option
 @report_file_option
@@ -238,22 +269,29 @@ def download_update(ctx: Context, database: Callable[..., Database], users: tupl
 @pass_context
 @docstring_format()
 def download_submissions(ctx: Context, database: Callable[..., Database], submission_id: tuple[int], replace: bool,
-                         dry_run: bool, verbose_report: bool, report_file: TextIO | None):
+                         retry: int | None, save_comments: bool, dry_run: bool, verbose_report: bool,
+                         report_file: TextIO | None):
     """
     Download single submissions, where {yellow}SUBMISSION_ID{reset} is the ID of the submission.
 
     If the {yellow}--replace{reset} option is used, database entries will be overwritten with new data (favorites will
     be maintained).
 
+    The {yellow}--retry{reset} option enables downloads retries for submission files and thumbnails up to 5 retries.
+
+    The {yellow}--save-comments{reset} option allows saving comments for downloaded submissions and journals. Comments
+    are otherwise ignored.
+
     The optional {yellow}--dry-run{reset} option disables downloading and saving and simply lists fetched entries
     """
     db: Database = database()
     api: FAAPI = open_api(db, ctx)
-    downloader: Downloader = Downloader(db, api, color=ctx.color, dry_run=dry_run)
+    downloader: Downloader = Downloader(db, api, color=ctx.color, comments=save_comments, retry=retry or 0,
+                                        replace=replace, dry_run=dry_run)
     if not dry_run:
         add_history(db, ctx, submission_id=submission_id, replace=replace)
     try:
-        downloader.download_submissions(list(submission_id), replace)
+        downloader.download_submissions(list(submission_id))
     except Unauthorized as err:
         secho(f"\nError: Unauthorized{(': ' + ' '.join(err.args)) if err.args else ''}", fg="red", color=ctx.color)
     except RequestException as err:
@@ -265,10 +303,12 @@ def download_submissions(ctx: Context, database: Callable[..., Database], submis
             downloader.verbose_report(report_file)
 
 
+# noinspection DuplicatedCode
 @download_app.command("journals", short_help="Download single journals.", no_args_is_help=True)
 @argument("journal_id", nargs=-1, required=True, type=IntRange(1),
           callback=lambda _c, _p, v: sorted(set(v), key=v.index))
-@option("--replace", is_flag=True, default=False, show_default=True, help="Replace submissions already in database.")
+@option("--replace", is_flag=True, default=False, show_default=True, help="Replace journals already in database.")
+@comments_option
 @dry_run_option
 @verbose_report_option
 @report_file_option
@@ -278,22 +318,26 @@ def download_submissions(ctx: Context, database: Callable[..., Database], submis
 @pass_context
 @docstring_format()
 def download_journals(ctx: Context, database: Callable[..., Database], journal_id: tuple[int], replace: bool,
-                      dry_run: bool, verbose_report: bool, report_file: TextIO | None):
+                      save_comments: bool, dry_run: bool, verbose_report: bool, report_file: TextIO | None):
     """
     Download single journals, where {yellow}JOURNAL_ID{reset} is the ID of the journal.
 
     If the {yellow}--replace{reset} option is used, database entries will be overwritten with new data (favorites will
     be maintained).
 
+    The {yellow}--save-comments{reset} option allows saving comments for downloaded submissions and journals. Comments
+    are otherwise ignored.
+
     The optional {yellow}--dry-run{reset} option disables downloading and saving and simply lists fetched entries.
     """
     db: Database = database()
     api: FAAPI = open_api(db, ctx)
-    downloader: Downloader = Downloader(db, api, color=ctx.color, dry_run=dry_run)
+    downloader: Downloader = Downloader(db, api, color=ctx.color, comments=save_comments, replace=replace,
+                                        dry_run=dry_run)
     if not dry_run:
         add_history(db, ctx, journal_id=journal_id, replace=replace)
     try:
-        downloader.download_journals(list(journal_id), replace)
+        downloader.download_journals(list(journal_id))
     except Unauthorized as err:
         secho(f"\nError: Unauthorized{(': ' + ' '.join(err.args)) if err.args else ''}", fg="red", color=ctx.color)
     except RequestException as err:
