@@ -11,19 +11,27 @@ from typing import TextIO
 from typing import TypeVar
 
 from click import echo
+from faapi import Comment
 from faapi import FAAPI
+from faapi import Journal
+from faapi import Submission
 from faapi import SubmissionPartial
 from faapi import UserPartial
+from faapi.comment import flatten_comments
 from faapi.exceptions import DisabledAccount
 from faapi.exceptions import NotFound
 from faapi.exceptions import NoticeMessage
 from faapi.exceptions import ServerError
+from faapi.journal import JournalPartial
 from falocalrepo_database import Column
 from falocalrepo_database import Database
 from falocalrepo_database.selector import SelectorBuilder as Sb
+from falocalrepo_database.tables import CommentsColumns
 from falocalrepo_database.tables import JournalsColumns
 from falocalrepo_database.tables import SubmissionsColumns
 from falocalrepo_database.tables import UsersColumns
+from falocalrepo_database.tables import journals_table
+from falocalrepo_database.tables import submissions_table
 from requests import RequestException
 from requests import Response
 
@@ -103,6 +111,18 @@ def download_catch(func: Callable[..., T], *args, **kwargs) -> tuple[T | None, i
         return None, 3
 
 
+def save_comments(db: Database, parent_table: str, parent_id: int, comments: list[Comment], *, replace: bool = False):
+    for comment in filter(lambda c: not c.hidden, flatten_comments(comments)):
+        db.comments.save_comment(
+            {CommentsColumns.ID.name: comment.id,
+             CommentsColumns.PARENT_TABLE.name: parent_table,
+             CommentsColumns.PARENT_ID.name: parent_id,
+             CommentsColumns.REPLY_TO.name: comment.reply_to.id if comment.reply_to else None,
+             CommentsColumns.AUTHOR.name: comment.author.name,
+             CommentsColumns.DATE.name: comment.date,
+             CommentsColumns.TEXT.name: comment.text}, replace=replace)
+
+
 class Bar:
     def __init__(self, length: int = 0, *, message: str = ""):
         self.length: int = length
@@ -145,10 +165,14 @@ class Bar:
 
 # noinspection DuplicatedCode
 class Downloader:
-    def __init__(self, db: Database, api: FAAPI, *, color: bool = True, dry_run: bool = False):
+    def __init__(self, db: Database, api: FAAPI, *, color: bool = True, retry: int = 0, comments: bool = False,
+                 replace: bool = False, dry_run: bool = False):
         self.db: Database = db
         self.output: OutputType = OutputType.rich if terminal_width() > 0 else OutputType.simple
         self.color: bool = color
+        self.retry: int = retry
+        self.save_comments: bool = comments
+        self.replace: bool = replace
         self.dry_run: bool = dry_run
         self.api: FAAPI = api
         self.bar_width: int = 10
@@ -302,7 +326,26 @@ class Downloader:
             self.bar_close(close_end)
         return err
 
-    def download_submission(self, submission_id: int, user_update: int, favorites: Iterable[str] | None,
+    def download_journal(self, journal_id: int, user_update: bool, replace: bool = False) -> int:
+        self.bar_clear()
+        self.bar_message("DOWNLOAD")
+        result, err = download_catch(self.api.journal, journal_id)
+        if self.err_to_bar(err):
+            self.journal_errors += [journal_id]
+            return err
+        journal: Journal = result
+        self.db.journals.save_journal({
+            **format_entry(dict(journal), self.db.journals.columns),
+            JournalsColumns.AUTHOR.name: journal.author.name,
+            JournalsColumns.USERUPDATE.name: user_update
+        }, replace=replace)
+        if self.save_comments:
+            save_comments(self.db, journals_table, journal.id, journal.comments, replace=replace)
+        self.db.commit()
+        self.bar_message("ADDED", green, always=True)
+        return 0
+
+    def download_submission(self, submission_id: int, user_update: bool, favorites: Iterable[str] | None,
                             thumbnail: str, replace: bool = False) -> int:
         self.bar_clear()
         self.bar_message("DOWNLOAD")
@@ -310,20 +353,32 @@ class Downloader:
         if self.err_to_bar(err):
             self.submission_errors += [submission_id]
             return err
-        submission, _ = result
+        submission: Submission = result[0]
         self.bar_clear()
         self.bar_close("\b")
         self.bar(7)
         file: bytes | None = self.download_bytes(submission.file_url)
+        retry: int = self.retry + 1
+        while file is None and (retry := retry - 1):
+            self.bar_message(f"RETRY {self.retry - retry + 1}", red)
+            self.api.handle_delay()
+            file = self.download_bytes(submission.file_url)
         self.bar_message(("#" * self.bar_width) if file else "ERROR", green if file else red, always=True)
         self.bar_close("]")
         self.bar(1)
         thumb: bytes | None = self.download_bytes(submission.thumbnail_url or thumbnail)
+        retry = self.retry + 1
+        while thumb is None and (retry := retry - 1):
+            self.bar_message(f"RETRY {self.retry - retry + 1}", red)
+            self.api.handle_delay()
+            thumb = self.download_bytes(submission.thumbnail_url or thumbnail)
         self.db.submissions.save_submission({**format_entry(dict(submission), self.db.submissions.columns),
-                                             "author": submission.author.name,
-                                             SubmissionsColumns.FAVORITE.value.name: {*favorites} if favorites else {},
-                                             SubmissionsColumns.USERUPDATE.value.name: int(user_update)},
+                                             SubmissionsColumns.AUTHOR.name: submission.author.name,
+                                             SubmissionsColumns.FAVORITE.name: {*favorites} if favorites else {},
+                                             SubmissionsColumns.USERUPDATE.name: user_update},
                                             file, thumb, replace=replace)
+        if self.save_comments:
+            save_comments(self.db, submissions_table, submission.id, submission.comments, replace=replace)
         self.db.commit()
         self.bar_message(("#" * self.bar_width) if thumb else "ERROR", green if thumb else red, always=True)
         self.bar_close()
@@ -336,7 +391,7 @@ class Downloader:
                              page_start: P, entry_id_getter: Callable[[T], int | str], entry_formats: tuple[str, str],
                              contains: Callable[[T], dict | None],
                              modify_checks: list[tuple[Callable[[T, dict], bool], str]],
-                             save: tuple[Callable[[T], int | None], str], stop: int = -1,
+                             save: tuple[Callable[[T, dict | None], int | None], str], stop: int = -1,
                              clear_last_found: bool = False, clear_found: bool = False,
                              ) -> tuple[int, tuple[list[int | str], list[int | str], list[int | str]]]:
         entries_added: list[int | str] = []
@@ -386,7 +441,8 @@ class Downloader:
                      nl=self.output == OutputType.simple, color=self.color)
                 self.bar()
                 self.bar_message("SEARCHING")
-                if curr_entry := contains(entry):
+                db_entry: dict | None = contains(entry)
+                if db_entry and not self.replace:
                     self.bar_message("IN DB", green, always=True)
                     if self.dry_run:
                         stop -= 1
@@ -396,7 +452,7 @@ class Downloader:
                     else:
                         modified: bool = False
                         for check, message in modify_checks:
-                            if modified := check(entry, curr_entry):
+                            if modified := check(entry, db_entry):
                                 self.db.commit()
                                 self.bar_message(message or "UPDATED", green, always=True)
                                 entries_modified.append(entry_id_getter(entry))
@@ -409,7 +465,7 @@ class Downloader:
                 elif self.dry_run:
                     self.bar_message("SKIPPED", green)
                 else:
-                    err = save[0](entry)
+                    err = save[0](entry, db_entry) or 0
                     if not self.err_to_bar(err) and save[1]:
                         self.bar_message(save[1], green, always=True)
                     entries_added.append(entry_id_getter(entry))
@@ -421,15 +477,24 @@ class Downloader:
         return 0, (entries_added, entries_modified, entries_errors)
 
     def download_user_journals(self, user: str, stop: int = -1, clear_last_found: bool = False) -> int:
+        def save(journal: JournalPartial, _db_entry: dict | None) -> int:
+            if self.save_comments:
+                return self.download_journal(journal.id, True, self.replace)
+            else:
+                self.db.journals.save_journal(
+                    {**format_entry(dict(journal), self.db.journals.columns),
+                     JournalsColumns.AUTHOR.name: journal.author.name,
+                     JournalsColumns.USERUPDATE.name: True},
+                    replace=self.replace)
+            return 0
+
         err, [entries_added, entries_modified, entries_errors] = self.download_user_folder(
             user=user, folder=Folder.journals, downloader_entries=self.api.journals, page_start=1,
             entry_id_getter=lambda j: j.id,
             entry_formats=("{0.id:010}", "{0.title}"),
             contains=lambda j: self.db.journals[j.id],
             modify_checks=[(lambda journal, _: self.db.journals.set_user_update(journal.id, True), "")],
-            save=(lambda journal: self.db.journals.save_journal(
-                {**format_entry(dict(journal) | {"author": journal.author.name}, self.db.journals.columns),
-                 JournalsColumns.USERUPDATE.value.name: 1, }), "ADDED"),
+            save=(save, "ADDED"),
             stop=stop, clear_last_found=clear_last_found
         )
         self.added_journals.extend(entries_added)
@@ -451,16 +516,22 @@ class Downloader:
                                                      self.db.submissions.set_folder(submission.id, folder.value)),
                               "UPDATED")]
 
+        def save(sub_partial: SubmissionPartial, db_entry: dict | None,) -> int:
+            return self.download_submission(
+                sub_partial.id,
+                (db_entry or {}).get(SubmissionsColumns.USERUPDATE.name, False) or folder != Folder.favorites,
+                [*(db_entry or {}).get(SubmissionsColumns.FAVORITE.name, []),
+                 *([user] if folder == Folder.favorites else [])],
+                sub_partial.thumbnail_url,
+                self.replace)
+
         err, [_entries_added, entries_modified, _entries_errors] = self.download_user_folder(
             user=user, folder=folder, downloader_entries=downloader, page_start=page_start,
             entry_id_getter=lambda s: s.id,
             entry_formats=("{0.id:010}", "{0.title}"),
             contains=lambda s: self.db.submissions[s.id],
             modify_checks=modify_checks,
-            save=(lambda sub_partial: self.download_submission(
-                sub_partial.id, int(folder != Folder.favorites),
-                [user] if folder == Folder.favorites else None,
-                sub_partial.thumbnail_url), ""),
+            save=(save, ""),
             stop=stop, clear_last_found=clear_last_found
         )
         self.modified_submissions.extend(entries_modified)
@@ -482,6 +553,7 @@ class Downloader:
                 return True
             return False
 
+        self.replace = False
         err, [entries_added, entries_modified, _entries_errors] = self.download_user_folder(
             user=user, folder=watchlist, downloader_entries=downloader, page_start=1,
             entry_id_getter=lambda u: u.name_url,
@@ -653,17 +725,19 @@ class Downloader:
         if not users_folders:
             return echo("No users to update")
 
+        self.replace = False
         self._download_users(users_folders, stop)
 
     # noinspection DuplicatedCode
-    def download_submissions(self, submission_ids: list[int], replace: bool = False):
+    def download_submissions(self, submission_ids: list[int]):
         header_width: int = (len(str(len(submission_ids))) * 2) + 2
         for i, submission_id in enumerate(submission_ids, 1):
             echo(f"{i}/{len(submission_ids)}".ljust(header_width) + f"{blue}{submission_id:010}{reset} ",
                  nl=self.output == OutputType.simple, color=self.color)
             self.bar()
             self.bar_message("SEARCHING")
-            if (entry := (self.db.submissions[submission_id] or {})) and not replace:
+            entry: dict = (self.db.submissions[submission_id] or {})
+            if entry and not self.replace:
                 self.bar_message("IN DB", green)
                 self.bar_close()
                 continue
@@ -672,19 +746,20 @@ class Downloader:
                 self.bar_close()
                 continue
             self.download_submission(submission_id,
-                                     entry.get(SubmissionsColumns.USERUPDATE.value.name, 0),
+                                     entry.get(SubmissionsColumns.USERUPDATE.value.name, False),
                                      entry.get(SubmissionsColumns.FAVORITE.value.name, {}),
-                                     "", replace)
+                                     "", self.replace)
 
     # noinspection DuplicatedCode
-    def download_journals(self, journal_ids: list[int], replace: bool = False):
+    def download_journals(self, journal_ids: list[int]):
         header_width: int = (len(str(len(journal_ids))) * 2) + 2
         for i, journal_id in enumerate(journal_ids, 1):
             echo(f"{i}/{len(journal_ids)}".ljust(header_width) + f"{blue}{journal_id:010}{reset} ",
                  nl=self.output == OutputType.simple, color=self.color)
             self.bar()
             self.bar_message("SEARCHING")
-            if (entry := (self.db.journals[journal_id] or {})) and not replace:
+            entry: dict = (self.db.journals[journal_id] or {})
+            if entry and not self.replace:
                 self.bar_message("IN DB", green)
                 self.bar_close()
                 continue
@@ -698,9 +773,11 @@ class Downloader:
                 continue
             self.db.journals.save_journal({
                 **format_entry(dict(journal), self.db.journals.columns),
-                "author": journal.author.name,
-                (u := JournalsColumns.USERUPDATE.value.name): entry.get(u, 0)
-            }, replace=replace)
+                JournalsColumns.AUTHOR.name: journal.author.name,
+                (u := JournalsColumns.USERUPDATE.name): entry.get(u, False)
+            }, replace=self.replace)
+            if self.save_comments:
+                save_comments(self.db, journals_table, journal.id, journal.comments, replace=self.replace)
             self.added_journals += [journal.id]
             self.db.commit()
             self.bar_message("ADDED", green, always=True)
